#ifndef HIPC_CLIENT_H
#define HIPC_CLIENT_H

#include <stddef.h>

#include <hipc.h>
#include <hipc/msg.h>

enum HIPC_member_type {
    HIPC_MBR_SIMPLE,            /* simple variable */
    HIPC_MBR_STRUCT,
    HIPC_MBR_ARRAY
};

/**
 * Machine-independent information of a member of a structure used in a client
 */
struct HIPC_cl_indp_mbr {
    enum HIPC_member_type type;
    char const *name;

    /* for HIPC_MEMBER_STRUCT */
    unsigned char struno;

    /* for HIPC_MEMBER_ARRAY */
    size_t ary_dim;
    size_t const *ary_max_idxs;
    enum HIPC_member_type elm_type;
    unsigned char elm_struno;   /* used when elem of array is struct */
};

/**
 * Machine-independent information of a structure used in a client
 */
struct HIPC_cl_indp_stru {
    unsigned int n_member;
    char const *name;
    struct HIPC_cl_indp_mbr const *mbr_p;
};

/**
 * Machine-independent information of a client
 */
struct HIPC_cl_indp {
    const unsigned char *cfgid;
    unsigned int cfgid_size;
    unsigned int n_sysmsg;
    unsigned int n_struct;
    struct HIPC_cl_indp_stru const *struct_p;
};

/**
 * Machine-dependent information of a structure used in a client
 */
struct HIPC_cl_dp_stru {
    unsigned char *offset_p;
    unsigned char *size_p;
};

/**
 * Machine-dependent information of a client
 */
struct HIPC_cl_dp {
    unsigned char endian;
    unsigned char const *struct_size_p;
    struct HIPC_cl_dp_stru *struct_p;
};

struct HIPC_ptr_link_list {
    struct HIPC_ptr_link_list *next;
    void *ptr;
};

struct HIPC_client {
    int endian_convert;
    struct HIPC_cl_indp const *indp_p;
    struct HIPC_cl_dp *cmdp_p; /**< Client Machine DePendent data */
    struct HIPC_cl_dp *smdp_p; /**< Server Machine DePendent data */
    unsigned char **simg_pp; /**< Server-side images of structures */
    unsigned char **cimg_pp; /**< Client-side images of structures */
    enum HIPC_errno (**hndlr_pp) (struct HIPC_client *, const hipc_msg,
                                  void *, void *);
    void **hdarg_pp; /**<  Handler's arguments */
    int std_errno;
    enum HIPC_errno hipc_errno;
    char err_detail_str[HIPC_IMG_ALLOCSIZE];
    struct HIPC_ptr_link_list *pllist_p;
};

#ifdef __cplusplus
extern "C" {
#endif

    extern void hipcClientSetStdErrno(struct HIPC_client *const cl_p);
    extern int hipcClientGetStdErrno(const struct HIPC_client *const cl_p);
    extern enum HIPC_errno
     hipcClientGetHipcErrno(const struct HIPC_client *const cl_p);
    extern const char *hipcClientGetErrStr(const struct HIPC_client *const
                                           cl_p);
    extern const char *hipcClientGetErrDtlStr(const struct HIPC_client
                                              *const cl_p);
    extern enum HIPC_errno
     hipcClientSetError(struct HIPC_client *const cl_p,
                        enum HIPC_errno hipc_errno, const char *const str);

    extern void hipcClientSetCimg(struct HIPC_client *const cl_p,
                                  const unsigned char struno,
                                  void *cimg_p);
    extern void
     hipcClientSetHandler(struct HIPC_client *const cl_p,
                          const unsigned char struno,
                          enum HIPC_errno (*hndlr_p) (struct HIPC_client *,
                                                      const hipc_msg,
                                                      void *, void *),
                          void *const arg);
    enum HIPC_errno
     hipcClientCallHandler(struct HIPC_client *const cl_p,
                           const unsigned char struno,
                           const hipc_msg msgR);

    extern unsigned char
     hipcClientGetNsysmsg(struct HIPC_client const *const cl_p);

    extern enum HIPC_errno
     hipcClientInit(struct HIPC_client *const cl_p,
                    struct HIPC_client const *const clbase_p);
    extern void hipcClientTerminate(struct HIPC_client *const cl_p);

    extern enum HIPC_errno
     hipcClientCheckMbrSize(struct HIPC_client *const cl_p);

    enum HIPC_errno
     hipcClientCheckStrutpl(struct HIPC_client *const cl_p,
                            const unsigned char struno,
                            const size_t offset, const size_t size);
    extern enum HIPC_errno
     hipcClientGetStrutpl(struct HIPC_client *const cl_p,
                          char const *const str,
                          unsigned char *struno_p,
                          size_t * offset_p, size_t * size_p);
    enum HIPC_errno
     hipcClientEachMbr(struct HIPC_client *const cl_p,
                       struct HIPC_cl_indp_mbr const *const mbr_p,
                       unsigned char *const begin,
                       unsigned char *const end,
                       unsigned char *const scur_p,
                       unsigned char *const ccur_p, const size_t ssize,
                       const size_t csize,
                       void (*fnc_p) (void *const ccur_p,
                                      void *const scur_p, const size_t n));
    extern enum HIPC_errno
     hipcClientUnpack(struct HIPC_client *const cl_p,
                      const unsigned char struno,
                      const size_t offset, const size_t size,
                      void *simg, void *cimg);
    extern enum HIPC_errno
     hipcClientPack(struct HIPC_client *const cl_p,
                    const unsigned char struno,
                    const size_t offset, const size_t size,
                    void *simg, void *cimg);

    enum HIPC_errno
     hipcClientFeedUnstrumsgAsError(struct HIPC_client *const cl_p,
                                    const hipc_msg msgR);
    extern enum HIPC_errno
     hipcClientFeedSysmsg(struct HIPC_client *const cl_p,
                          const hipc_msg msgR);
    extern enum HIPC_errno
     hipcClientFeedStrumsg(struct HIPC_client *const cl_p,
                           const hipc_msg msgR);

    extern void
     hipcGenmsgHello(hipc_msg msg, const struct HIPC_client *const cl_p);
    extern enum HIPC_errno
     hipcGenmsgGet(hipc_msg msg, struct HIPC_client *const cl_p,
                   const unsigned char struno,
                   const size_t offset, const size_t size);
    extern enum HIPC_errno
     hipcGenmsgPut(hipc_msg msg, struct HIPC_client *const cl_p,
                   const unsigned char struno,
                   const size_t offset, const size_t size);
    extern void hipcGenmsgBye(hipc_msg msg);

#ifdef __cplusplus
}
#endif
#endif
