#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "client.c"

static void print_pllist(FILE * fp,
                         struct HIPC_ptr_link_list const *pllist_p)
{
    if (NULL == fp) {
        return;
    }

    while (NULL != pllist_p) {
        fprintf(stderr, "p %p/%p\n", (void *) pllist_p,
                (void *) pllist_p->ptr);
        pllist_p = pllist_p->next;
    }

    return;
}

/**
 * Check normal termination when there is no memory allocation.
 */
static void test_client_memope_no_allocation(void)
{
    struct HIPC_client cl;

    cl.pllist_p = NULL;
    print_pllist(stdout, cl.pllist_p);
    hipcClientTerminate(&cl);
    print_pllist(stdout, cl.pllist_p);
    assert(NULL == cl.pllist_p);

    return;
}

/**
 * Check normal termination when there is one malloc.
 */
static void test_client_memope_malloc(void)
{
    struct HIPC_client cl;
    void *tmp_p;

    cl.pllist_p = NULL;
    tmp_p = hipcClientMalloc(&cl, 5);
    assert(NULL != cl.pllist_p);
    assert(tmp_p == cl.pllist_p->ptr);
    assert(NULL == cl.pllist_p->next);
    print_pllist(stdout, cl.pllist_p);

    hipcClientTerminate(&cl);
    assert(NULL == cl.pllist_p);
    print_pllist(stdout, cl.pllist_p);

    return;
}

/**
 * Check normal termination when there is one calloc.
 */
static void test_client_memope_calloc(void)
{
    struct HIPC_client cl;
    unsigned char *tmp_p;
    int i;

    cl.pllist_p = NULL;
    tmp_p = hipcClientCalloc(&cl, 5, 7);
    for (i = 0; i < (5 * 7); ++i) {
        assert(0 == tmp_p[i]);
    }
    assert(NULL != cl.pllist_p);
    assert(tmp_p == cl.pllist_p->ptr);
    assert(NULL == cl.pllist_p->next);
    print_pllist(stdout, cl.pllist_p);

    hipcClientTerminate(&cl);
    assert(NULL == cl.pllist_p);
    print_pllist(stdout, cl.pllist_p);

    return;
}

/**
 * Check normal termination when there are multiple memory allocations.
 */
static void test_client_memope_multi(void)
{
    struct HIPC_client cl;
    void *pary_mem[3];
    struct HIPC_ptr_link_list const *pary_pllist[3];

    cl.pllist_p = NULL;
    pary_mem[0] = hipcClientMalloc(&cl, 128);
    assert(NULL != cl.pllist_p);
    pary_pllist[0] = cl.pllist_p;
    assert(pary_pllist[0]->ptr == pary_mem[0]);
    assert(NULL == pary_pllist[0]->next);
    print_pllist(stdout, cl.pllist_p);
    fprintf(stdout, "\n");

    pary_mem[1] = hipcClientCalloc(&cl, 8, 64);
    assert(NULL != cl.pllist_p);
    pary_pllist[1] = cl.pllist_p;
    assert(pary_pllist[1]->ptr == pary_mem[1]);
    assert(pary_pllist[0] == pary_pllist[1]->next);
    assert(pary_pllist[0]->ptr == pary_mem[0]);
    assert(NULL == pary_pllist[0]->next);
    print_pllist(stdout, cl.pllist_p);
    fprintf(stdout, "\n");

    pary_mem[2] = hipcClientMalloc(&cl, 256);
    assert(NULL != cl.pllist_p);
    pary_pllist[2] = cl.pllist_p;
    assert(pary_pllist[2]->ptr == pary_mem[2]);
    assert(pary_pllist[1] == pary_pllist[2]->next);
    assert(pary_pllist[1]->ptr == pary_mem[1]);
    assert(pary_pllist[0] == pary_pllist[1]->next);
    assert(pary_pllist[0]->ptr == pary_mem[0]);
    assert(NULL == pary_pllist[0]->next);
    print_pllist(stdout, cl.pllist_p);
    fprintf(stdout, "\n");

    hipcClientTerminate(&cl);
    assert(NULL == cl.pllist_p);
    print_pllist(stdout, cl.pllist_p);

    return;
}

int main(int argc, char *argv[])
{
    if (2 != argc) {
        fprintf(stderr, "Usage: %s test_id\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    switch (argv[1][0]) {
    case 'a':
        test_client_memope_no_allocation();
        break;
    case 'b':
        test_client_memope_malloc();
        break;
    case 'c':
        test_client_memope_calloc();
        break;
    case 'd':
        test_client_memope_multi();
        break;
    default:
        fprintf(stderr, "Invalid test_id: %s \n", argv[1]);
        exit(EXIT_FAILURE);
    }

    return 0;
}
